###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

Floodlight_Report_Dimensions_Schema = [{
    'name': 'Activity',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Activity_Group',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Activity_Group_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Activity_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Activity_Date_Time',
    'type': 'DATETIME',
    'mode': 'NULLABLE'
}, {
    'name': 'Ad',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Ad_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Ad_Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Advertiser',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Advertiser_Group',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Advertiser_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Asset',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Asset_Category',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Asset_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Asset_Orientation',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Audience_Targeted',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Browser_Platform',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Campaign',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Campaign_End_Date',
    'type': 'DATE',
    'mode': 'NULLABLE'
}, {
    'name': 'Campaign_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Campaign_Start_Date',
    'type': 'DATE',
    'mode': 'NULLABLE'
}, {
    'name': 'Channel_Mix',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'City',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Click_Count',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Click_Through_Url',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Connection_Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Conversion_Referrer',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Conversion_Url',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Country',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Groups_1',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Groups_2',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Pixel_Size',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Creative_Version',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Rich_Media_Custom_Event_Count',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Rich_Media_Custom_Event_Path_Summary',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Date',
    'type': 'DATE',
    'mode': 'NULLABLE'
}, {
    'name': 'Days_Since_Attributed_Interaction',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Days_Since_First_Interaction',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Designated_Market_Area_Dma',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_1',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_1_Field_Value_1',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_1_Field_Value_2',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_1_Field_Value_3',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_1_Value',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_1_Value_Id',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_2',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_2_Field_Value_1',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_2_Field_Value_2',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_2_Field_Value_3',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_2_Value',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_2_Value_Id',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_3',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_3_Field_Value_1',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_3_Field_Value_2',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_3_Field_Value_3',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_3_Value',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_3_Value_Id',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_4',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_4_Field_Value_1',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_4_Field_Value_2',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_4_Field_Value_3',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_4_Value',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_4_Value_Id',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_5',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_5_Field_Value_1',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_5_Field_Value_2',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_5_Field_Value_3',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_5_Value',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Element_5_Value_Id',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Profile',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Dynamic_Profile_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Attribution_Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Configuration',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_1',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_2',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_3',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_4',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_5',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_6',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_7',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_8',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_9',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_10',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_11',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_12',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_13',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_14',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_15',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_16',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_17',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_18',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_19',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_20',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_21',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_22',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_23',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_24',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_25',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_26',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_27',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_28',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_29',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_30',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_31',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_32',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_33',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_34',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_35',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_36',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_37',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_38',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_39',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_40',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_41',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_42',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_43',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_44',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_45',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_46',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_47',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_48',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_49',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_50',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_51',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_52',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_53',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_54',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_55',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_56',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_57',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_58',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_59',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_60',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_61',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_62',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_63',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_64',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_65',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_66',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_67',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_68',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_69',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_70',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_71',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_72',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_73',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_74',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_75',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_76',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_77',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_78',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_79',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_80',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_81',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_82',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_83',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_84',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_85',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_86',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_87',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_88',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_89',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_90',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_91',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_92',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_93',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_94',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_95',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_96',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_97',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_98',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_99',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Floodlight_Variable_100',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Has_Backup_Image',
    'type': 'BOOLEAN',
    'mode': 'NULLABLE'
}, {
    'name': 'Has_Counters',
    'type': 'BOOLEAN',
    'mode': 'NULLABLE'
}, {
    'name': 'Has_Exits',
    'type': 'BOOLEAN',
    'mode': 'NULLABLE'
}, {
    'name': 'Has_Timers',
    'type': 'BOOLEAN',
    'mode': 'NULLABLE'
}, {
    'name': 'Has_Dynamic_Impressions',
    'type': 'BOOLEAN',
    'mode': 'NULLABLE'
}, {
    'name': 'Has_Expansions',
    'type': 'BOOLEAN',
    'mode': 'NULLABLE'
}, {
    'name': 'Has_Full_Screen_Impressions',
    'type': 'BOOLEAN',
    'mode': 'NULLABLE'
}, {
    'name': 'Has_Full_Screen_Video_Completions',
    'type': 'BOOLEAN',
    'mode': 'NULLABLE'
}, {
    'name': 'Has_Full_Screen_Video_Plays',
    'type': 'BOOLEAN',
    'mode': 'NULLABLE'
}, {
    'name': 'Has_Full_Screen_Views',
    'type': 'BOOLEAN',
    'mode': 'NULLABLE'
}, {
    'name': 'Has_Html5_Impressions',
    'type': 'BOOLEAN',
    'mode': 'NULLABLE'
}, {
    'name': 'Has_Interactive_Impressions',
    'type': 'BOOLEAN',
    'mode': 'NULLABLE'
}, {
    'name': 'Has_Manual_Closes',
    'type': 'BOOLEAN',
    'mode': 'NULLABLE'
}, {
    'name': 'Has_Video_Companion_Clicks',
    'type': 'BOOLEAN',
    'mode': 'NULLABLE'
}, {
    'name': 'Has_Video_Completions',
    'type': 'BOOLEAN',
    'mode': 'NULLABLE'
}, {
    'name': 'Has_Video_First_Quartile_Completions',
    'type': 'BOOLEAN',
    'mode': 'NULLABLE'
}, {
    'name': 'Has_Video_Full_Screen',
    'type': 'BOOLEAN',
    'mode': 'NULLABLE'
}, {
    'name': 'Has_Video_Interactions',
    'type': 'BOOLEAN',
    'mode': 'NULLABLE'
}, {
    'name': 'Has_Video_Midpoints',
    'type': 'BOOLEAN',
    'mode': 'NULLABLE'
}, {
    'name': 'Has_Video_Mutes',
    'type': 'BOOLEAN',
    'mode': 'NULLABLE'
}, {
    'name': 'Has_Video_Pauses',
    'type': 'BOOLEAN',
    'mode': 'NULLABLE'
}, {
    'name': 'Has_Video_Plays',
    'type': 'BOOLEAN',
    'mode': 'NULLABLE'
}, {
    'name': 'Has_Video_Progress_Events',
    'type': 'BOOLEAN',
    'mode': 'NULLABLE'
}, {
    'name': 'Has_Video_Replays',
    'type': 'BOOLEAN',
    'mode': 'NULLABLE'
}, {
    'name': 'Has_Video_Skips',
    'type': 'BOOLEAN',
    'mode': 'NULLABLE'
}, {
    'name': 'Has_Video_Stops',
    'type': 'BOOLEAN',
    'mode': 'NULLABLE'
}, {
    'name': 'Has_Video_Third_Quartile_Completions',
    'type': 'BOOLEAN',
    'mode': 'NULLABLE'
}, {
    'name': 'Has_Video_Unmutes',
    'type': 'BOOLEAN',
    'mode': 'NULLABLE'
}, {
    'name': 'Has_Video_Views',
    'type': 'BOOLEAN',
    'mode': 'NULLABLE'
}, {
    'name': 'Hour',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Hours_Since_Attributed_Interaction',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Hours_Since_First_Interaction',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Impression_Count',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Interaction_Channel',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Interaction_Count_Click_Tracker',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Interaction_Count_Mobile_Rich_Media',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Interaction_Count_Mobile_Static_Image',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Interaction_Count_Mobile_Video',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Interaction_Count_Natural_Search',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Interaction_Count_Paid_Search',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Interaction_Count_Rich_Media',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Interaction_Count_Static_Image',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Interaction_Count_Video',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Interaction_Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Mobile_Carrier',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Month',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Natural_Search_Engine_Country',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Natural_Search_Engine_Property',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Natural_Search_Engine_Url',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Natural_Search_Landing_Page',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Natural_Search_Landing_Page_Query_String',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Natural_Search_Processed_Landing_Page',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Natural_Search_Processed_Landing_Page_Query_String',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Natural_Search_Query',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Num_Value',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Operating_System',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Operating_System_Version',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Ord_Value',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Package_Roadblock',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Package_Roadblock_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Package_Roadblock_Strategy',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Ad',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Ad_Group',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Ad_Group_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Ad_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Advertiser',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Advertiser_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Agency',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Agency_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Bid_Strategy',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Bid_Strategy_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Campaign',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Campaign_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Engine_Account',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Engine_Account_Category',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Engine_Account_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_External_Ad_Group_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_External_Ad_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_External_Campaign_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_External_Keyword_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Keyword',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Keyword_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Labels',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Landing_Page_Url',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Legacy_Keyword_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Paid_Search_Match_Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Path_Length',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Path_Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement_End_Date',
    'type': 'DATE',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement_Pixel_Size',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Placement_Start_Date',
    'type': 'DATE',
    'mode': 'NULLABLE'
}, {
    'name': 'Platform_Type',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Rendering_Id',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Video_Length',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Site_Dcm',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Site_Site_Directory',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Site_Id_Site_Directory',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Site_Id_Dcm',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Site_Keyname',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Rich_Media_Standard_Event_Count',
    'type': 'INTEGER',
    'mode': 'NULLABLE'
}, {
    'name': 'Rich_Media_Standard_Event_Path_Summary',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'State_Region',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Tran_Value',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'U_Value',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Week',
    'type': 'STRING',
    'mode': 'NULLABLE'
}, {
    'name': 'Zip_Postal_Code',
    'type': 'STRING',
    'mode': 'NULLABLE'
}]



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

# SEE: https://github.com/geduldig/TwitterAPI
# SEE: https://developer.twitter.com/en/docs/basics/rate-limits

from time import sleep
from TwitterAPI import TwitterAPI

from starthinker.util.project import project
from starthinker.util.data import get_rows, put_rows

# FOR WOEID SEE
# http://cagricelebi.com/blog/dear-twitter-please-stop-using-woeid/
# https://archive.org/details/geoplanet_data_7.10.0.zip
# https://github.com/Ray-SunR/woeid
# https://stackoverflow.com/questions/12434591/get-woeid-from-city-name

TWITTER_API = None


def get_twitter_api():
  global TWITTER_API
  if TWITTER_API is None:
    TWITTER_API = TwitterAPI(
        project.task['key'], project.task['secret'], auth_type='oAuth2')
  return TWITTER_API


TWITTER_TRENDS_PLACE_SCHEMA = [
    {
        'name': 'Woeid',
        'type': 'INTEGER'
    },
    {
        'name': 'Name',
        'type': 'STRING'
    },
    {
        'name': 'Url',
        'type': 'STRING'
    },
    {
        'name': 'Promoted_Content',
        'type': 'STRING',
        'mode': 'NULLABLE'
    },
    {
        'name': 'Query',
        'type': 'STRING',
    },
    {
        'name': 'Tweet_Volume',
        'type': 'INTEGER'
    },
]


def twitter_trends_places():
  if project.verbose:
    print('TWITTER TRENDS PLACE')
  print('PL',
        list(get_rows(project.task['auth'], project.task['trends']['places'])))

  for place in get_rows(project.task['auth'], project.task['trends']['places']):
    if project.verbose:
      print('PLACE', place)
    results = get_twitter_api().request('trends/place', {'id': int(place)})
    for r in results:
      if project.verbose:
        print('RESULT', r['name'])
      yield [
          place, r['name'], r['url'], r['promoted_content'], r['query'],
          r['tweet_volume']
      ]
    print('.', end='')
    sleep(15 * 60 / 75)  # rate limit ( improve to retry )


TWITTER_TRENDS_CLOSEST_SCHEMA = [
    {
        'name': 'Latitude',
        'type': 'FLOAT'
    },
    {
        'name': 'Longitude',
        'type': 'FLOAT'
    },
    {
        'name': 'Country',
        'type': 'STRING'
    },
    {
        'name': 'Country_Code',
        'type': 'STRING'
    },
    {
        'name': 'Name',
        'type': 'STRING',
        'mode': 'NULLABLE'
    },
    {
        'name': 'Parent_Id',
        'type': 'INTEGER',
    },
    {
        'name': 'Place_Type_Code',
        'type': 'INTEGER'
    },
    {
        'name': 'Place_Type_Name',
        'type': 'STRING'
    },
    {
        'name': 'URL',
        'type': 'STRING'
    },
    {
        'name': 'Woeid',
        'type': 'INTEGER'
    },
]


def twitter_trends_closest():
  if project.verbose:
    print('TWITTER TRENDS CLOSEST')
  for row in get_rows(project.task['auth'], project.task['trends']['closest']):
    lat, lon = row[0], row[1]
    results = api.request('trends/closest', {'lat': lat, 'long': lon})
    for r in results:
      yield [
          lat, lon, r['country'], r['countryCode'], r['name'], r['parentid'],
          r['placeType']['code'], r['placeType']['name'], r['url'], r['woeid']
      ]


TWITTER_TRENDS_AVAILABLE_SCHEMA = TWITTER_TRENDS_CLOSEST_SCHEMA


def twitter_trends_available():
  if project.verbose:
    print('TWITTER TRENDS AVAILABLE')
  results = api.request('trends/available', {})
  for r in results:
    yield [
        r['country'], r['countryCode'], r['name'], r['parentid'],
        r['placeType']['code'], r['placeType']['name'], r['url'], r['woeid']
    ]


@project.from_parameters
def twitter():
  if project.verbose:
    print('TWITTER')

  rows = None

  if 'trends' in project.task:
    if 'places' in project.task['trends']:
      rows = twitter_trends_places()
      project.task['out']['bigquery']['schema'] = TWITTER_TRENDS_PLACE_SCHEMA
      project.task['out']['bigquery']['skip_rows'] = 0
    elif 'closest' in project.task['trends']:
      rows = twitter_trends_closest()
      project.task['out']['bigquery']['schema'] = TWITTER_TRENDS_CLOSEST_SCHEMA
      project.task['out']['bigquery']['skip_rows'] = 0
    else:
      rows = twitter_trends_available()
      project.task['out']['bigquery'][
          'schema'] = TWITTER_TRENDS_AVAILABLE_SCHEMA
      project.task['out']['bigquery']['skip_rows'] = 0

  if rows:
    put_rows(project.task['auth'], project.task['out'], rows)


if __name__ == '__main__':
  twitter()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import json

from django.core.management.base import BaseCommand, CommandError
from django.conf import settings

from starthinker_ui.recipe.models import Recipe


class Command(BaseCommand):
  help = 'Moves database recipes to json files or pub/sub topic'

  def add_arguments(self, parser):
    parser.add_argument(
        '--remote',
        action='store_true',
        dest='remote',
        default=False,
        help='Run jobs remotely using pub sub.',
    )

    parser.add_argument(
        '--recipe',
        action='store',
        dest='recipe',
        default=None,
        help='Run a specific recipe.',
    )

    parser.add_argument(
        '--force',
        action='store_true',
        dest='force',
        default=False,
        help='Force execution regardless of schedule.',
    )

  def handle(self, *args, **kwargs):

    for recipe in (Recipe.objects.filter(pk=kwargs['recipe'])
                   if kwargs['recipe'] else Recipe.objects.filter(active=True)):
      try:
        if kwargs['remote']:
          print('Dispatch: %s' % recipe.uid())
          if kwargs['force']:
            recipe.force()
        elif settings.UI_CRON:
          print('Write: %s/recipe_%d.json' % (settings.UI_CRON, recipe.pk))
          with open(settings.UI_CRON + '/recipe_%d.json' % recipe.pk, 'w') as f:
            f.write(json.dumps(recipe.get_json()))
        else:
          raise Exception('Neither UI_CRON configured nor remote specified.')

      except (KeyboardInterrupt, SystemExit):
        raise

      except Exception as e:
        print('DEPLOY ERROR:', str(e))



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""Handler that executes { "dcm":{...}} task in recipe JSON.

This script translates JSON instructions into operations on DCM reporting.
It deletes, or creates, and/or downloads DCM reports.  See JSON files in
this directory for examples of operations.

This script uses put_rows as defined in util/data/README.md. This allows
multiple destinations for downloaded reports. To add a destination modify
the util/data/__init__.py functions.

Note

The underlying libraries use streaming download buffers, no disk is used.
Buffers are controlled in config.py.
For superusers, this script will use the internal API, bypassing the
need for profiles.
Reports uploaded to BigQuery use automatic schema detection based on official
proto files.

"""

from starthinker.util.project import project
from starthinker.util.data import get_rows, put_rows
from starthinker.util.dcm import report_delete, report_filter, report_build, report_file, report_to_rows, report_clean, report_schema, report_run


@project.from_parameters
def dcm():
  if project.verbose:
    print('DCM')

  # stores existing report json
  report = None

  # check if report is to be deleted
  if project.task.get('delete', False):
    if project.verbose:
      print(
          'DCM DELETE', project.task['report'].get('name', None) or
          project.task['report'].get('body', {}).get('name', None) or
          project.task['report'].get('report_id', None))
    report_delete(
        project.task['auth'],
        project.task['report']['account'],
        project.task['report'].get('report_id', None),
        project.task['report'].get('name', None) or
        project.task['report'].get('body', {}).get('name', None),
    )

  # check if report is to be run
  if project.task.get('report_run_only', False):
    if project.verbose:
      print(
          'DCM REPORT RUN', project.task['report'].get('name', None) or
          project.task['report'].get('report_id', None))
    report_run(
        project.task['auth'],
        project.task['report']['account'],
        project.task['report'].get('report_id', None),
        project.task['report'].get('name', None),
    )

  # check if report is to be created
  if 'body' in project.task['report']:
    if project.verbose:
      print('DCM BUILD', project.task['report']['body']['name'])

    if 'filters' in project.task['report']:
      project.task['report']['body'] = report_filter(
          project.task['auth'], project.task['report']['body'],
          project.task['report']['filters'])

    report = report_build(
        project.task['auth'], project.task['report']['body'].get('accountId') or
        project.task['report']['account'], project.task['report']['body'])

  # moving a report
  if 'out' in project.task:
    filename, report = report_file(
        project.task['auth'],
        project.task['report']['account'],
        project.task['report'].get('report_id', None),
        project.task['report'].get('name', None) or
        project.task['report'].get('body', {}).get('name', None),
        project.task['report'].get('timeout', 10),
    )

    if report:
      if project.verbose:
        print('DCM FILE', filename)

      # clean up the report
      rows = report_to_rows(report)
      rows = report_clean(rows)

      # if bigquery, remove header and determine schema
      schema = None
      if 'bigquery' in project.task['out']:
        schema = report_schema(next(rows))
        project.task['out']['bigquery']['schema'] = schema
        project.task['out']['bigquery']['skip_rows'] = 0

      # write rows using standard out block in json ( allows customization across all scripts )
      if rows:
        put_rows(project.task['auth'], project.task['out'], rows)


if __name__ == '__main__':
  dcm()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import os
import fcntl
import json
import math
import uuid
import time
import signal
import traceback
import subprocess
import threading
from datetime import datetime

from django.core.management.base import BaseCommand, CommandError
from django.db import connection, transaction
from django.conf import settings

from starthinker_ui.recipe.models import Recipe, utc_milliseconds, utc_milliseconds_to_timezone, JOB_LOOKBACK_MS, JOB_INTERVAL_MS
from starthinker_ui.recipe.log import log_manager_start, log_manager_end, log_manager_scale, log_manager_timeout, log_manager_error
from starthinker_ui.recipe.log import log_job_timeout, log_job_error, log_job_start, log_job_end, log_job_cancel
from starthinker_ui.recipe.log import log_verbose, get_instance_name
from starthinker_ui.recipe.compute import group_instances_delete

MANAGER_ON = True
MANAGER_HEALTHY = True
IDLE_INTERVAL = 5 * 60  # if worker is idle for 5 minutes, shut it down


def signal_exit(self, signum):
  global MANAGER_ON
  MANAGER_ON = False


signal.signal(signal.SIGINT, signal_exit)
signal.signal(signal.SIGTERM, signal_exit)


def worker_ping(worker_uid, recipe_uids):
  # update recipes that belong to this worker
  if recipe_uids:
    Recipe.objects.filter(
        worker_uid=worker_uid,
        id__in=recipe_uids).update(worker_utm=utc_milliseconds())


def worker_status(worker_uid, recipe_uid, script, instance, hour, event, stdout,
                  stderr):
  try:
    job = Recipe.objects.get(worker_uid=worker_uid, id=recipe_uid)
    job.set_task(script, instance, hour, event, stdout, stderr)
  except Recipe.DoesNotExist:
    print('Expired Worker Job:', worker_uid, recipe_uid, script, instance, hour,
          event)


def worker_pull(worker_uid, jobs=1):
  """Atomic reservation of worker in jobs.

  Args: - worker_uid ( string ) - identifies a unique worker, must be same for
  every call from same worker. - jobs ( integer ) - number of jobs to pull
  """

  jobs_all = []
  jobs_new = []

  worker_utm = utc_milliseconds()
  worker_lookback = worker_utm - JOB_LOOKBACK_MS

  if jobs:

    with transaction.atomic():

      # find recipes that are available but have not been pinged recently from all workers ( pulls from pool )
      where = Recipe.objects.filter(
          active=True,
          worker_utm__lte=worker_lookback,
          job_utm__lte=worker_utm,
      ).exclude(job_utm=0).select_for_update(
          skip_locked=True).order_by('worker_utm').values_list(
              'id', flat=True)[:jobs]

      # mark those recipes as belonging to this worker
      Recipe.objects.filter(id__in=where).update(
          worker_uid=worker_uid, worker_utm=worker_utm)

  # find all recipes that belong to this worker and check if they have new tasks
  for job in Recipe.objects.filter(active=True, worker_uid=worker_uid):
    jobs_all.append(job.id)
    task = job.get_task()  # also resets status
    if job.worker_utm == worker_utm and task:  # jobs with current timestamp are new ( odds of a ping matching this worker_utm? ), isolate evens and odds?
      jobs_new.append(task)

  return jobs_all, jobs_new


def worker_downscale():
  name = get_instance_name()
  if name != 'UNKNOWN':
    group_instances_delete(name)


def make_non_blocking(file_io):
  fd = file_io.fileno()
  fl = fcntl.fcntl(fd, fcntl.F_GETFL)
  fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)


class Workers():

  def __init__(self, uid, jobs_maximum, timeout_seconds, trace=False):
    self.uid = uid or get_instance_name()
    self.timeout_seconds = timeout_seconds
    self.trace = trace
    self.jobs_maximum = jobs_maximum
    self.jobs = []

    self.lock_thread = threading.Lock()
    self.ping_event = threading.Event()
    self.ping_thread = threading.Thread(target=self.ping)
    self.ping_thread.start()

    self.busy_time = datetime.utcnow()

  def available(self):
    return self.jobs_maximum - len(self.jobs)

  def pull(self):

    self.lock_thread.acquire()

    # get jobs for this worker ( threadsafe and takes a long time, so outside of thread lock )
    jobs_all, jobs_new = worker_pull(self.uid, jobs=self.available())

    # remove all lost jobs from ping
    jobs_remove = []
    last_job = len(self.jobs)
    while last_job > 0:
      last_job -= 1
      if self.jobs[last_job]['recipe']['setup']['uuid'] not in jobs_all:
        jobs_remove.append(self.jobs[last_job])
        del self.jobs[last_job]

    # add all new jobs to ping
    self.jobs.extend(jobs_new)

    # allow pings to resume with up to date list
    self.lock_thread.release()

    # shut down all removed jobs
    try:
      for job in jobs_remove:
        if job.get('job', {}).get('process'):
          job['job']['process'].kill()
        self.cleanup(job)
        log_job_cancel(job)
    except Exception as e:
      log_manager_error(traceback.format_exc())

  def run(self, job, force=False):

    job['recipe']['setup'].setdefault('timeout_seconds', self.timeout_seconds)

    job['job'] = {
        'worker': self.uid,
        'id': str(uuid.uuid1()),
        'process': None,
        'utc': datetime.utcnow(),
    }

    filename = '%s/%s.json' % (settings.UI_CRON, job['job']['id'])

    with open(filename, 'w') as job_file:
      job_file.write(json.dumps(job['recipe'], default=str))

    command = [
        '%s/starthinker_virtualenv/bin/python' % settings.UI_ROOT,
        '-u',
        '-W',
        'ignore',
        '%s/starthinker/task/%s/run.py' % (settings.UI_ROOT, job['script']),
        filename,
        '-i',
        str(job['instance']),
        '--verbose',
    ]

    if self.trace:
      command.append('--trace_file')

    job['job']['process'] = subprocess.Popen(
        command,
        shell=False,
        cwd=settings.UI_ROOT,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE)

    make_non_blocking(job['job']['process'].stdout)
    make_non_blocking(job['job']['process'].stderr)

  def cleanup(self, job):
    filename = '%s/%s.json' % (settings.UI_CRON, job['job']['id'])
    if os.path.exists(filename):
      os.remove(filename)

  def ping(self):
    global MANAGER_HEALTHY
    while MANAGER_HEALTHY and not self.ping_event.wait(JOB_INTERVAL_MS / 1000):
      self.lock_thread.acquire()
      try:
        worker_ping(self.uid,
                    [job['recipe']['setup']['uuid'] for job in self.jobs])
      except Exception as e:
        log_manager_error(traceback.format_exc())
        MANAGER_HEALTHY = False
      self.lock_thread.release()

  def poll(self):

    for job in self.jobs:

      # if job changes state, this is set, then sent to database
      status = None

      # start job if it is not already running
      if 'job' not in job:

        self.run(job)
        log_job_start(job)
        status = 'JOB_START'
        stdout = None
        stderr = None

      # if already running check status
      else:

        # read any incremental data from the process ( made non-blocking at construction )
        stdout = job['job']['process'].stdout.read()
        if stdout is not None:
          stdout = stdout.decode()
        stderr = job['job']['process'].stderr.read()
        if stderr is not None:
          stderr = stderr.decode()

        # if process still running, check timeout or ping keep alive
        poll = job['job']['process'].poll()
        if poll is None:

          # check if task is a timeout
          if (datetime.utcnow() - job['job']['utc']
             ).total_seconds() > job['recipe']['setup']['timeout_seconds']:
            status = 'JOB_TIMEOUT'
            job['job']['process'].kill()
            self.cleanup(job)
            log_job_timeout(job)
            job['job']['process'] = None

          # otherwise task is running, update stdout and stderr if present
          elif stdout or stderr:
            status = 'JOB_START'

        # if process has return code, check if task is complete or error
        else:
          self.cleanup(job)

          # if error scrap whole worker and flag error
          if poll != 0:
            status = 'JOB_ERROR'
            log_job_error(job)
            job['job']['process'] = None

          # if success, pop task off the stack and flag success
          else:
            status = 'JOB_END'
            log_job_end(job)
            job['job']['process'] = None

      # if status is set, send it to the database
      if status:
        worker_status(job['job']['worker'], job['recipe']['setup']['uuid'],
                      job['script'], job['instance'], job['hour'], status,
                      stdout, stderr)

    # remove all workers without a process, they are done
    if self.jobs:
      self.lock_thread.acquire()
      self.jobs = [
          job for job in self.jobs if job['job']['process'] is not None
      ]
      self.lock_thread.release()

    # update the busy time if jobs exist
    if len(self.jobs) != 0:
      self.busy_time = datetime.utcnow()

    # if workers remain, return True
    return bool(self.jobs)

  def idle(self):
    return (datetime.utcnow() - self.busy_time).seconds > IDLE_INTERVAL

  def shutdown(self):
    # wait for jobs to finish
    while self.poll():
      time.sleep(JOB_INTERVAL_MS / 1000)

    # turn off threads ( ping )
    self.ping_event.set()


class Command(BaseCommand):
  help = 'Executes a recipe job and writes status to the databse.'

  def add_arguments(self, parser):
    parser.add_argument(
        '--worker',
        action='store',
        dest='worker',
        default='',
        help='Name of worker to use when requesting jobs.',
    )

    parser.add_argument(
        '--jobs',
        action='store',
        dest='jobs',
        default=3,
        type=int,
        help='Maximum number of jobs simlutanelous processes to start within this worker.',
    )

    parser.add_argument(
        '--timeout',
        action='store',
        dest='timeout',
        default=60 * 60 * 12,  # 12 hours
        type=int,
        help='Default seconds to allow a task to run before timing it out, also controlled by recipe.',
    )

    parser.add_argument(
        '--verbose',
        action='store_true',
        dest='verbose',
        default=False,
        help='Causes log messages to also print.',
    )

    parser.add_argument(
        '--trace',
        action='store_true',
        dest='trace',
        default=False,
        help='Create an execution trace in /tmp/starthinker_trace.log.',
    )

    parser.add_argument(
        '--test',
        action='store_true',
        dest='test',
        default=False,
        help='Set test mode to execute loop only once and return workers.',
    )

  def handle(self, *args, **kwargs):
    global MANAGER_ON
    global MANAGER_HEALTHY

    MANAGER_ON = True
    MANAGER_HEALTHY = True

    if kwargs['test']:
      print('Starting Up...')

    if kwargs['verbose']:
      log_verbose()

    log_manager_start()

    if kwargs['test']:
      print('Initializing Workers...')

    workers = Workers(
        kwargs['worker'],
        kwargs['jobs'],
        kwargs['timeout'],
        kwargs['trace'],
    )

    try:

      while MANAGER_HEALTHY and MANAGER_ON:

        # load jobs
        workers.pull()

        time.sleep(JOB_INTERVAL_MS / 1000)

        # evaluate jobs
        workers.poll()

        # check if worker needs to scale down
        if workers.idle():
          MANAGER_ON = False
          log_manager_timeout()
        else:
          time.sleep(JOB_INTERVAL_MS / 1000)

        if kwargs['test']:
          MANAGER_ON = False

    except KeyboardInterrupt:
      MANAGER_ON = False

    except Exception as e:
      if kwargs['test']:
        print(str(e))
      log_manager_error(traceback.format_exc())

    if MANAGER_HEALTHY:
      if kwargs['test']:
        print('Shutting Down...')
      workers.shutdown()

    log_manager_end()

    # worker will terminate itself in a group safe way
    worker_downscale()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker.util.bigquery import table_create
from starthinker.util.data import get_rows
from starthinker.util.data import put_rows
from starthinker.util.google_api import API_DV360
from starthinker.util.google_api.discovery_to_bigquery import Discovery_To_BigQuery
from starthinker.util.project import project
from starthinker.util.regexp import lookup_id
from starthinker.util.sheets import sheets_clear

from starthinker.task.dv_editor.patch import patch_clear
from starthinker.task.dv_editor.patch import patch_log
from starthinker.task.dv_editor.patch import patch_masks
from starthinker.task.dv_editor.patch import patch_preview


def campaign_clear():
  table_create(
      project.task['auth_bigquery'],
      project.id,
      project.task['dataset'],
      'DV_Campaigns',
      Discovery_To_BigQuery('displayvideo',
                            'v1').method_schema('advertisers.campaigns.list'),
  )

  sheets_clear(project.task['auth_sheets'], project.task['sheet'], 'Campaigns', 'B2:Z')


def campaign_load():

  # load multiple partners from user defined sheet
  def campaign_load_multiple():
    rows = get_rows(
        project.task['auth_sheets'], {
            'sheets': {
                'sheet': project.task['sheet'],
                'tab': 'Advertisers',
                'range': 'A2:A'
            }
        })

    for row in rows:
      yield from API_DV360(
          project.task['auth_dv'], iterate=True).advertisers().campaigns().list(
              advertiserId=lookup_id(row[0])).execute()

  # write campaigns to database and sheet
  put_rows(
      project.task['auth_bigquery'], {
          'bigquery': {
              'dataset':
                  project.task['dataset'],
              'table':
                  'DV_Campaigns',
              'schema':
                  Discovery_To_BigQuery(
                      'displayvideo',
                      'v1').method_schema('advertisers.campaigns.list'),
              'format':
                  'JSON'
          }
      }, campaign_load_multiple())

  # write campaigns to sheet
  rows = get_rows(
      project.task['auth_bigquery'], {
          'bigquery': {
              'dataset':
                  project.task['dataset'],
              'query':
                  """SELECT
         CONCAT(P.displayName, ' - ', P.partnerId),
         CONCAT(A.displayName, ' - ', A.advertiserId),
         CONCAT(C.displayName, ' - ', C.campaignId),
         C.entityStatus
         FROM `{dataset}.DV_Campaigns` AS C
         LEFT JOIN `{dataset}.DV_Advertisers` AS A
         ON C.advertiserId=A.advertiserId
         LEFT JOIN `{dataset}.DV_Partners` AS P
         ON A.partnerId=P.partnerId
       """.format(**project.task),
              'legacy':
                  False
          }
      })

  put_rows(project.task['auth_sheets'], {
      'sheets': {
          'sheet': project.task['sheet'],
          'tab': 'Campaigns',
          'range': 'B2'
      }
  }, rows)


def campaign_commit(patches):
  for patch in patches:
    if not patch.get('campaign'):
      continue
    print('API CALL', patch['action'], patch['advertiser'], patch['campaign'])
    try:
      if patch['action'] == 'DELETE':
        response = API_DV360(
            project.task['auth_dv']).advertisers().campaigns().delete(
                **patch['parameters']).execute()
        patch['success'] = response
      elif patch['action'] == 'PATCH':
        response = API_DV360(project.task['auth_dv']).advertisers().campaigns().patch(
            **patch['parameters']).execute()
        patch['success'] = response['campaignId']
    except Exception as e:
      patch['error'] = str(e)
    finally:
      patch_log(patch)
  patch_log()



URI = """{
 "auth": {
  "oauth2": {
   "scopes": {
    "https://www.googleapis.com/auth/userinfo.email": {
     "description": "View your email address"
    }
   }
  }
 },
 "basePath": "/_ah/api/snippets/v1/",
 "baseUrl": "https://snippets-hrdb.googleplex.com/_ah/api/snippets/v1/",
 "batchPath": "batch",
 "description": "Snippets API Library",
 "discoveryVersion": "v1",
 "icons": {
  "x16": "https://www.gstatic.com/images/branding/product/1x/googleg_16dp.png",
  "x32": "https://www.gstatic.com/images/branding/product/1x/googleg_32dp.png"
 },
 "id": "snippets:v1",
 "kind": "discovery#restDescription",
 "methods": {
  "listByConstraints": {
   "httpMethod": "GET",
   "id": "snippets.listByConstraints",
   "parameters": {
    "startDate": {
     "location": "query",
     "type": "string"
    },
    "endDate": {
     "location": "query",
     "type": "string"
    },
    "startWeekNumber": {
     "format": "int32",
     "location": "query",
     "type": "integer"
    },
    "endWeekNumber": {
     "format": "int32",
     "location": "query",
     "type": "integer"
    },
    "quarterNumber": {
     "format": "int32",
     "location": "query",
     "type": "integer"
    },
    "yearNumber": {
     "format": "int32",
     "location": "query",
     "type": "integer"
    },
    "userGroupKey": {
     "location": "query",
     "type": "string"
    },
    "includeDrafts": {
     "location": "query",
     "type": "boolean"
    }
   },
   "path": "snippets",
   "response": {
    "$ref": "SnippetEntityCollection"
   },
   "scopes": [
    "https://www.googleapis.com/auth/snippets",
    "https://www.googleapis.com/auth/snippets.readonly"
   ]
  }
 },
 "name": "snippets",
 "parameters": {
  "alt": {
   "default": "json",
   "description": "Data format for the response.",
   "enum": [
    "json"
   ],
   "enumDescriptions": [
    "Responses with Content-Type of application/json"
   ],
   "location": "query",
   "type": "string"
  },
  "fields": {
   "description": "Selector specifying which fields to include in a partial response.",
   "location": "query",
   "type": "string"
  },
  "key": {
   "description": "API key. Your API key identifies your project and provides you with API access, quota, and reports. Required unless you provide an OAuth 2.0 token.",
   "location": "query",
   "type": "string"
  },
  "oauth_token": {
   "description": "OAuth 2.0 token for the current user.",
   "location": "query",
   "type": "string"
  },
  "prettyPrint": {
   "default": "true",
   "description": "Returns response with indentations and line breaks.",
   "location": "query",
   "type": "boolean"
  },
  "quotaUser": {
   "description": "Available to use for quota purposes for server-side applications. Can be any arbitrary string assigned to a user, but should not exceed 40 characters. Overrides userIp if both are provided.",
   "location": "query",
   "type": "string"
  },
  "userIp": {
   "description": "IP address of the site where the request originates. Use this if you want to enforce per-user limits.",
   "location": "query",
   "type": "string"
  }
 },
 "protocol": "rest",
 "rootUrl": "https://snippets-hrdb.googleplex.com/_ah/api/",
 "schemas": {
  "Key": {
   "id": "Key",
   "properties": {
    "appId": {
     "type": "string"
    },
    "complete": {
     "type": "boolean"
    },
    "id": {
     "format": "int64",
     "type": "string"
    },
    "kind": {
     "type": "string"
    },
    "name": {
     "type": "string"
    },
    "namespace": {
     "type": "string"
    },
    "parent": {
     "$ref": "Key"
    }
   },
   "type": "object"
  },
  "SnippetEntity": {
   "id": "SnippetEntity",
   "properties": {
    "deleted": {
     "type": "boolean"
    },
    "draft": {
     "type": "boolean"
    },
    "key": {
     "$ref": "Key"
    },
    "keyNameOrId": {
     "type": "string"
    },
    "markdown": {
     "type": "boolean"
    },
    "modifiedDate": {
     "format": "date-time",
     "type": "string"
    },
    "personId": {
     "format": "int64",
     "type": "string"
    },
    "quarter": {
     "format": "int32",
     "type": "integer"
    },
    "submitDate": {
     "format": "date-time",
     "type": "string"
    },
    "text": {
     "$ref": "Text"
    },
    "textFromString": {
     "type": "string"
    },
    "visibility": {
     "type": "string"
    },
    "weekNumber": {
     "format": "int32",
     "type": "integer"
    },
    "year": {
     "format": "int32",
     "type": "integer"
    }
   },
   "type": "object"
  },
  "SnippetEntityCollection": {
   "id": "SnippetEntityCollection",
   "properties": {
    "items": {
     "items": {
      "$ref": "SnippetEntity"
     },
     "type": "array"
    }
   },
   "type": "object"
  },
  "Text": {
   "id": "Text",
   "properties": {
    "value": {
     "type": "string"
    }
   },
   "type": "object"
  }
 },
 "servicePath": "snippets/v1/",
 "version": "v1"
}
"""



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

#SEE: https://pymotw.com/2/sys/tracing.html

import sys

from starthinker.config import TRACE_FILE

STARTHINKER_TRACE_TO_FILE = False
STARTHINKER_TRACE_TO_PRINT = False


def is_starthinker_module(frame):
  if '__file__' in frame.f_globals:
    module_path = frame.f_globals['__file__']
    return 'starthinker' in module_path and 'starthinker_virtualenv' not in module_path
  else:
    return False


def starthinker_trace_log(entry):
  global STARTHINKER_TRACE_TO_PRINT
  global STARTHINKER_TRACE_TO_FILE

  if STARTHINKER_TRACE_TO_PRINT:
    print(entry)

  if STARTHINKER_TRACE_TO_FILE:
    with open(TRACE_FILE, 'a+') as log_file:
      log_file.write('%s\n' % entry)


def starthinker_trace(frame, event, arg, returns=True):

  co = frame.f_code
  func_name = co.co_name
  line_no = frame.f_lineno
  filename = co.co_filename

  if event == 'exception':
    exc_type, exc_value, exc_traceback = arg
    starthinker_trace_log(
        'Exception: %s on line %s of %s with type %s and value %s' %
        (func_name, line_no, filename, exc_type.__name__, exc_value))
    return

  elif is_starthinker_module(frame):

    if event == 'call':
      starthinker_trace_log('Call To: %s on line %s of %s' %
                            (func_name, line_no, filename))
      return starthinker_trace

    elif event == 'return' and returns:
      starthinker_trace_log('Return From: %s on line %s of %s with value %s' %
                            (func_name, line_no, filename, arg))
      return

  return


def starthinker_trace_start(to_print=True, to_file=False):
  global STARTHINKER_TRACE_TO_PRINT
  global STARTHINKER_TRACE_TO_FILE

  STARTHINKER_TRACE_TO_PRINT = to_print
  STARTHINKER_TRACE_TO_FILE = to_file

  if STARTHINKER_TRACE_TO_PRINT or STARTHINKER_TRACE_TO_FILE:
    sys.settrace(starthinker_trace)


if __name__ == '__main__':

  def c():
    raise RuntimeError('exception message goes here')

  def b():
    try:
      c()
    except Exception as e:
      pass
    return 'response_from_b '

  def a():
    val = b()
    return val.replace('_b', '_a')

  from starthinker.util.regexp import date_to_str

  starthinker_trace_start()
  a()

  date_to_str(None)



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

'''
--------------------------------------------------------------

Before running this Airflow module...

  Install StarThinker in cloud composer from open source:

    pip install git+https://github.com/google/starthinker

  Or push local code to the cloud composer plugins directory:

    source install/deploy.sh
    4) Composer Menu
    l) Install All

--------------------------------------------------------------

CM To BigQuery IO Synch

Migrate CM Placements to DV360 IO.

Fill in the CM accont information.
Fill in the DV360 accont information.
Click the 'Save' button to save configuration.
Click the 'Run Now' button to create the workflow.
Visit <b>BigQuery->UNDEFINED->UNDEFINED->DV360_IO_REVIEW</b> to check for updates.
Visit <b>BigQuery->UNDEFINED->UNDEFINED->DV360_IO_INSERT</b> to approve the updates.

'''

from starthinker_airflow.factory import DAG_Factory

# Add the following credentials to your Airflow configuration.
USER_CONN_ID = "starthinker_user" # The connection to use for user authentication.
GCP_CONN_ID = "starthinker_service" # The connection to use for service authentication.

INPUTS = {
  'recipe_slug': '',  # Place where tables will be created in BigQuery.
  'dv360_advertisers': '',  # Comma seperated.
  'cm_account': '',  # Comma seperated.
  'cm_advertiser': '',  # Comma seperated.
  'dv360_advertiser': '',  # Comma seperated.
  'dv360_campaign': '',  # Comma seperated.
}

TASKS = [
  {
    'dataset': {
      'auth': 'service',
      'dataset': {
        'field': {
          'name': 'recipe_slug',
          'kind': 'string',
          'description': 'Place where tables will be created in BigQuery.'
        }
      }
    }
  },
  {
    'sheets': {
      'auth': 'user',
      'sheet': 'BB Demo',
      'tab': 'Rules',
      'range': 'A1:C',
      'header': True,
      'out': {
        'auth': 'service',
        'bigquery': {
          'dataset': {
            'field': {
              'name': 'recipe_slug',
              'kind': 'string',
              'description': 'Place where tables will be created in BigQuery.'
            }
          },
          'table': 'Rules'
        }
      }
    }
  },
  {
    'dv360_api': {
      'auth': 'user',
      'endpoints': [
        'advertisers.insertionOrders'
      ],
      'advertisers': {
        'single_cell': True,
        'values': {
          'field': {
            'name': 'dv360_advertisers',
            'kind': 'integer_list',
            'description': 'Comma seperated.'
          }
        }
      },
      'out': {
        'auth': 'service',
        'dataset': {
          'field': {
            'name': 'recipe_slug',
            'kind': 'string',
            'description': 'Place where tables will be created in BigQuery.'
          }
        }
      }
    }
  },
  {
    'dcm_api': {
      'auth': 'user',
      'endpoints': [
        'campaigns',
        'placements',
        'placementGroups'
      ],
      'accounts': {
        'values': {
          'field': {
            'name': 'cm_account',
            'kind': 'integer_list',
            'description': 'Comma seperated.'
          }
        }
      },
      'advertisers': {
        'values': {
          'field': {
            'name': 'cm_advertiser',
            'kind': 'integer_list',
            'description': 'Comma seperated.'
          }
        }
      },
      'out': {
        'auth': 'service',
        'dataset': {
          'field': {
            'name': 'recipe_slug',
            'kind': 'string',
            'description': 'Place where tables will be created in BigQuery.'
          }
        }
      }
    }
  },
  {
    'bigquery': {
      'auth': 'service',
      'from': {
        'legacy': False,
        'parameters': [
          {
            'field': {
              'name': 'dv360_advertiser',
              'kind': 'integer_list',
              'description': 'Comma seperated.'
            }
          },
          {
            'field': {
              'name': 'dv360_campaign',
              'kind': 'integer_list',
              'description': 'Comma seperated.'
            }
          },
          {
            'field': {
              'name': 'recipe_slug',
              'kind': 'string',
              'description': 'Place where tables will be written in BigQuery.'
            }
          },
          {
            'field': {
              'name': 'recipe_slug',
              'kind': 'string',
              'description': 'Place where tables will be written in BigQuery.'
            }
          },
          {
            'field': {
              'name': 'recipe_slug',
              'kind': 'string',
              'description': 'Place where tables will be written in BigQuery.'
            }
          }
        ],
        'query': "SELECT   [PARAMETER] AS advertiserId,   [PARAMETER] AS campaignId,   REGEXP_REPLACE(CM_P.name, '_1X1.*', '') AS displayName,   'ENTITY_STATUS_DRAFT' AS entityStatus,   STRUCT(     'PACING_PERIOD_FLIGHT' AS pacingPeriod,     'PACING_TYPE_AHEAD' AS pacingType   ) AS pacing,   STRUCT(     false AS unlimited,     'TIME_UNIT_DAYS' AS timeUnit,     1 AS timeUnitCount,     CASE       WHEN CM_P.name LIKE '%FF%' AND CM_P.name LIKE '%PZN%'THEN 6       WHEN CM_P.name LIKE '%FF%' THEN 3       WHEN CM_P.name LIKE '%HOL%' THEN 6       WHEN CM_P.name LIKE '%BTC%' THEN 3     END AS maxImpressions   ) AS frequencyCap,   STRUCT(     'PERFORMANCE_GOAL_TYPE_CPM' AS performanceGoalType,     '10' AS performanceGoalAmountMicros   ) AS performanceGoal,   STRUCT(     'BUDGET_UNIT_CURRENCY' AS budgetUnit,     'INSERTION_ORDER_AUTOMATION_TYPE_BUDGET' AS automationType,     [       STRUCT(        '10000' AS budgetAmountMicros,        '' AS description,        STRUCT (          STRUCT (            EXTRACT(YEAR FROM GREATEST(CM_C.startDate, CURRENT_DATE())) AS year,            EXTRACT(MONTH FROM GREATEST(CM_C.startDate, CURRENT_DATE())) AS month,            EXTRACT(DAY FROM GREATEST(CM_C.startDate, CURRENT_DATE())) AS day          ) AS startDate,          STRUCT (            EXTRACT(YEAR FROM CM_C.endDate) AS year,            EXTRACT(MONTH FROM CM_C.endDate) AS month,            EXTRACT(DAY FROM CM_C.endDate) AS day          ) AS endDate        ) AS dateRange       )     ] AS budgetSegments   ) AS budget,   STRUCT(     STRUCT(       '10' AS bidAmountMicros     ) AS fixedBid   ) AS bidStrategy FROM `[PARAMETER].CM_placements` As CM_P LEFT JOIN `[PARAMETER].CM_campaigns` As CM_C ON CM_P.campaignId=CM_C.id LEFT JOIN `[PARAMETER].CM_placementGroups` As CM_G ON CM_P.placementGroupId=CM_G.id WHERE CM_P.advertiserID=4461789 AND CM_P.name LIKE 'PKG%' /*AND CM_G.placementGroupType = 'PACKAGE' */ ORDER BY displayName ;"
      },
      'to': {
        'dataset': {
          'field': {
            'name': 'recipe_slug',
            'kind': 'string',
            'description': 'Place where tables will be created in BigQuery.'
          }
        },
        'view': 'DV360_IO_LOGIC'
      }
    }
  },
  {
    'bigquery': {
      'auth': 'service',
      'from': {
        'legacy': False,
        'parameters': [
          {
            'field': {
              'name': 'recipe_slug',
              'kind': 'string',
              'description': 'Place where tables will be written in BigQuery.'
            }
          }
        ],
        'query': 'SELECT STRUCT(   ROW_NUMBER() OVER() AS NUMBER,   displayName IN (SELECT displayName FROM `[PARAMETER].DV360_advertisers_insertionOrders`) AS DUPLICATE ) AS PREVIEW, * FROM `[PARAMETER].DV360_IO_LOGIC` ORDER BY displayName ;'
      },
      'to': {
        'dataset': {
          'field': {
            'name': 'recipe_slug',
            'kind': 'string',
            'description': 'Place where tables will be created in BigQuery.'
          }
        },
        'view': 'DV360_IO_PREVIEW'
      }
    }
  },
  {
    'bigquery': {
      'auth': 'service',
      'from': {
        'legacy': False,
        'parameters': [
          {
            'field': {
              'name': 'recipe_slug',
              'kind': 'string',
              'description': 'Place where tables will be written in BigQuery.'
            }
          }
        ],
        'query': ' SELECT * EXCEPT(PREVIEW) FROM `[PARAMETER].DV360_IO_PREVIEW` WHERE PREVIEW.NUMBER IN(0,0,0,0,0) AND PREVIEW.DUPLICATE=False LIMIT 1 ;'
      },
      'to': {
        'dataset': {
          'field': {
            'name': 'recipe_slug',
            'kind': 'string',
            'description': 'Place where tables will be created in BigQuery.'
          }
        },
        'view': 'DV360_IO_INSERT'
      }
    }
  },
  {
    'dv360_api': {
      'auth': 'user',
      'bigquery': {
        'auth': 'service',
        'dataset': {
          'field': {
            'name': 'recipe_slug',
            'kind': 'string',
            'description': 'Place where tables will be created in BigQuery.'
          }
        },
        'as_object': True,
        'table': 'DV360_IO_INSERT'
      },
      'insert': 'advertisers.insertionOrders'
    }
  }
]

DAG_FACTORY = DAG_Factory('cm_to_dv360_io', { 'tasks':TASKS }, INPUTS)
DAG_FACTORY.apply_credentails(USER_CONN_ID, GCP_CONN_ID)
DAG = DAG_FACTORY.execute()

if __name__ == "__main__":
  DAG_FACTORY.print_commandline()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker_ui.recipe.scripts import Script
from starthinker.script.parse import json_get_fields, dict_to_string, fields_to_string

AIRFLOW_TEMPLATE = """###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

\'\'\'
--------------------------------------------------------------

Before running this Airflow module...

  Install StarThinker in cloud composer ( recommended ):

    From Release: pip install starthinker
    From Open Source: pip install git+https://github.com/google/starthinker

  Or push local code to the cloud composer plugins directory ( if pushing local code changes ):

    source install/deploy.sh
    4) Composer Menu
    l) Install All

--------------------------------------------------------------

  If any recipe task has "auth" set to "user" add user credentials:

    1. Ensure an RECIPE['setup']['auth']['user'] = [User Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_user", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/deploy_commandline.md#optional-setup-user-credentials

--------------------------------------------------------------

  If any recipe task has "auth" set to "service" add service credentials:

    1. Ensure an RECIPE['setup']['auth']['service'] = [Service Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_service", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/cloud_service.md

--------------------------------------------------------------

{title}

{description}

{instructions}

--------------------------------------------------------------

This StarThinker DAG can be extended with any additional tasks from the following sources:
  - https://google.github.io/starthinker/
  - https://github.com/google/starthinker/tree/master/dags

\'\'\'

from starthinker.airflow.factory import DAG_Factory

INPUTS = {inputs}

RECIPE = {recipe}

DAG_FACTORY = DAG_Factory('{dag}', RECIPE, INPUTS)
DAG = DAG_FACTORY.generate()

if __name__ == "__main__":
  DAG_FACTORY.print_commandline()
"""

def script_to_dag(dag_name,
                  title,
                  description,
                  instructions,
                  script,
                  parameters={}):
  return AIRFLOW_TEMPLATE.format(**{
    'title':title,
    'description':description,
    'instructions':'  - ' + '\n  - '.join(instructions),
    'inputs':fields_to_string(
      json_get_fields(script),
      parameters
    ),
    'recipe':dict_to_string(
      script,
      skip=('fields',)
    ),
    'dag':dag_name
  })



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from operator import itemgetter
from urllib.parse import urlencode

from starthinker.util.project import project
from starthinker.util.csv import rows_to_csv
from starthinker.util.sheets import sheets_read
from starthinker.util.bigquery import query_to_rows
from starthinker.util.email import send_email
from starthinker.util.email.template import EmailTemplate

# OWNER SCHEMA
#Account Name - STRING
#Account Owner - STRING
#Account Owner Email - STRING
#DCM Network ID - INTEGER
#DBM Partner ID - INTEGER
#DS Account ID - INTEGER
#Studio Account ID - INTEGER

# SOLUTION SCHEMA
# Account ID - INTEGER
# Variant ID - INTEGER  ( can be same as account id if not used )
# VARIANT - STRING ( title added to list to allow subaccounts, for example advertisers in DBM )
# Score - FLOAT ( Percent: 100 = Client Already Executing Well ( low need ), 1 = Client Currentnot Executing Well ( high need ) )

# IMPACT SCHEMA
# Variant ID - INTEGER ( must match column name above )
# Impact - PERCENT

IMPACT_SCHEMA = [
    {
        'name': 'Impact',
        'type': 'STRING'
    },
    {
        'name': 'Level',
        'type': 'INTEGER'
    },
]

OFFER_SCHEMA = [
    {
        'name': 'Client',
        'type': 'STRING'
    },
    {
        'name': 'Coverage',
        'type': 'INTEGER'
    },
    {
        'name': 'Request',
        'type': 'STRING'
    },
]


def get_owners():
  if project.verbose:
    print('GETTING OWNERS')

  owners = []

  if 'sheet' in project.task['owners']:
    owners = sheets_read(project.task['auth'],
                         project.task['owners']['sheet']['url'],
                         project.task['owners']['sheet']['tab'],
                         project.task['owners']['sheet']['range'])
  elif 'bigquery' in project.task['owners']:
    owners = query_to_rows(project.task['auth'], project.id,
                           project.task['owners']['bigquery']['dataset'],
                           project.task['owners']['bigquery']['query'])

  # group account owners by email, create easy lookup sets for ids
  owners_grouped = {}
  for owner in owners:
    try:

      owners_grouped.setdefault(
          owner[2], {
              'Account Name': owner[0],
              'Account Owner': owner[1],
              'Account Email': owner[2],
              'DCM Network ID': [],
              'DBM Partner ID': [],
              'DS Account ID': [],
              'Studio Account ID': [],
          })

      if len(owner) > 3 and owner[3]:
        owners_grouped[owner[2]]['DCM Network ID'].append(str(owner[3]))
      if len(owner) > 4 and owner[4]:
        owners_grouped[owner[2]]['DBM Partner ID'].append(str(owner[4]))
      if len(owner) > 5 and owner[5]:
        owners_grouped[owner[2]]['DS Account ID'].append(str(owner[5]))
      if len(owner) > 6 and owner[6]:
        owners_grouped[owner[2]]['Studio Account ID'].append(str(owner[6]))

    except IndexError:
      print('ERROR:', owner)
      pass

  if project.verbose:
    print('GOT OWNERS:', len(owners))

  return owners_grouped.values()


def get_impacts():
  if project.verbose:
    print('GETTING IMPACTS')

  impacts = []

  if 'sheet' in project.task['impacts']:
    impacts = sheets_read(project.task['auth'],
                          project.task['impacts']['sheet']['url'],
                          project.task['impacts']['sheet']['tab'],
                          project.task['impacts']['sheet']['range'])
  elif 'bigquery' in project.task['impacts']:
    impacts = query_to_rows(project.task['auth'], project.id,
                            project.task['impacts']['bigquery']['dataset'],
                            project.task['impacts']['bigquery']['query'])

  # for easy lookup use dictionary
  impacts = dict([(str(i[0]), float(i[1])) for i in impacts])

  if project.verbose:
    print('GOT IMPACTS:', len(impacts))

  return impacts


def get_solutions():
  if project.verbose:
    print('GETTING SCORES')

  for solution in project.task['solutions']:
    scores = []

    if 'sheet' in solution:
      scores = sheets_read(project.task['auth'], solution['sheet']['url'],
                           solution['sheet']['tab'], solution['sheet']['range'])
    elif 'bigquery' in solution:
      scores = query_to_rows(project.task['auth'], project.id,
                             solution['bigquery']['dataset'],
                             solution['bigquery']['query'])

    # for easy lookup use dictionary
    solution['scores'] = {}
    for score in scores:
      solution['scores'].setdefault(str(score[0]), [])
      solution['scores'][str(score[0])].append({
          'variant_id': str(score[1]),
          'variant': score[2],
          'score': float(score[3])
      })

    if project.verbose:
      print('GOT SCORES:', len(solution['scores']))

  return project.task['solutions']


def compose_link(link, parameters):
  return '%s%s%s' % (link, '&' if '?' in link else '?', urlencode(parameters))


def compose_email_solution_centric(owner):
  if project.verbose:
    print('COMPOSING: ', owner['Account Email'])

  # start an email template
  email = EmailTemplate()
  email.greeting(owner['Account Owner'])
  email.paragraph(project.task['email']['introduction'])

  # loop through solutions
  rows = []
  for solution in owner['Solutions']:
    email.header('Your %d Biggest %s Impact Opportunities This Week' %
                 (project.task['offers'], solution['Solution']['name']))

    # create offer matrix for each solution
    email.table(
        OFFER_SCHEMA,
        [
            (
                offer['Variant'],
                '%d%%' % int(offer['Score'] * 100),
                compose_link(
                    solution['Solution']['link'],
                    {
                        'solution':
                            solution['Solution']['name'],
                        'requester':
                            owner['Account Email'],
                        'name':
                            offer['Variant'],
                        'dbm':
                            '%s:%s' % (offer['Account_ID'], offer['Variant_ID'])
                            if solution['Solution']['key'] == 'DBM Partner ID'
                            else '',  # comma seperated list
                        'dcm':
                            '%s:%s' % (offer['Account_ID'], offer['Variant_ID'])
                            if solution['Solution']['key'] == 'DCM Network ID'
                            else '',  # comma seperated list
                        'ds':
                            '%s:%s' % (offer['Account_ID'], offer['Variant_ID'])
                            if solution['Solution']['key'] == 'DS Account ID'
                            else '',  # comma seperated list
                        'sa':
                            '%s:%s' %
                            (offer['Account_ID'], offer['Variant_ID']) if
                            solution['Solution']['key'] == 'Studio Account ID'
                            else '',  # comma seperated list
                    })) for offer in solution['Offers']
        ])

    email.paragraph(
        "Click the request link to schedule a solution for that client.  This does not deploy a solution to the client, it only contacts a specialist indicating you'd like to evaluate this solution for this client."
    )

    email.header('About ' + solution['Solution']['name'])
    email.image(solution['Solution']['image'], solution['Solution']['sample'])
    email.paragraph(solution['Solution']['description'])

    email.paragraph('Product Gap')
    email.paragraph(solution['Solution']['gap'])

    # solution pitch
    email.paragraph('Benefits')
    email.list(solution['Solution']['pitches'])

    # solution impact
    email.table(IMPACT_SCHEMA, [
        (i[0], '%d%%' % i[1]) for i in solution['Solution']['impacts'].items()
    ])

    email.button('Learn More', project.task['link'], big=True)

  #print(email.get_html())

  send_email('user', owner['Account Email'], project.task['email']['from'],
             project.task['email']['cc'], project.task['email']['subject'],
             email.get_text(), email.get_html())


def assemble_offers_solution_centric():

  owners, solutions, impacts = get_owners(), get_solutions(), get_impacts()

  # produce the following table
  # owner [
  #   solution - ( detials ) [
  #    [ client - variant, score, impact ] ( sorted by impact )

  if project.verbose:
    print('ASSEMBLING OFFERS')

  count = 0
  for owner in owners:
    owner.setdefault('Solutions', [])

    # for each owner and solution assemble a list of offers
    for solution in solutions:
      offers = []

      for account_id in owner[solution['key']]:

        scores = solution['scores'].get(account_id, [])

        # if account id has a solution score add it to the offers
        for score in scores:

          offers.append({
              'Account_ID':
                  account_id,
              'Variant_ID':
                  score['variant_id'],
              'Variant':
                  score['variant'],
              'Score':
                  score['score'],
              'Impact':
                  impacts.get(score['variant_id'], 1) *
                  (1 - score['score']
                  )  # Impact is how much revenue can be improved
          })
          count += 1

      # if offers for this account and solution exist, keep only the top ones ( largest impact to lowest )
      if offers:
        offers.sort(key=itemgetter('Impact'), reverse=True)
        offers = offers[:project.task['offers']]
        owner['Solutions'].append({'Solution': solution, 'Offers': offers})

  if project.verbose:
    print('ASSEMBLED OFFERS', count)

  if project.verbose:
    print('SENDING OFFERS')

  # send emails
  count = 0
  for owner in owners:
    if owner['Solutions']:
      compose_email_solution_centric(owner)
      count += 1
      if count == 10:
        exit()

  if project.verbose:
    print('SENT OFFERS', count)


@project.from_parameters
def marketing():
  assemble_offers_solution_centric()


if __name__ == '__main__':
  marketing()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################


class StringExtensions:

  @staticmethod
  def convertDateStrToDateTimeStr(date, time='00:00:00'):
    """Convert Date string (YYYY-MM-DD) to a datetime string by adding the desired time (YYYY-MM-DDTHH:mm:SSZ)

                Args:
                        date: the date as a string to be converted
                        time: the time as a string to be added to the date

                Returns:
                        A string representation of a datetime in the following
                        format YYYY-MM-DDTHH:mm:SSZ
                """
    if not date == None:
      date = '%sT%sZ' % (date, time)

    return date

  @staticmethod
  def convertDateTimeStrToDateStr(datetime):
    """  Convert a DateTime string (YYYY-MM-DDTHH:mm:SSZ) to just a Date string by removing the time (YYYY-MM-DD)

                Args:
                        datetime: the datetime as a string

                Returns:
                        A string representation of the date in the following
                        format YYYY-MM-DD
                """
    if not datetime == None and 'T' in datetime:
      datetime = datetime.split('T')[0]

    return datetime



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import argparse
import subprocess
import sys
import textwrap

from starthinker.config import UI_ROOT
from starthinker.util.project import is_scheduled
from starthinker.util.project import get_project

EXIT_ERROR = 1
EXIT_SUCCESS = 0

def main():

  # this is a helper function, these inputs mirror util.project.Project singleton used by tasks because they are pass through to each task
  parser = argparse.ArgumentParser(
      formatter_class=argparse.RawDescriptionHelpFormatter,
      description=textwrap.dedent("""\
      Command line to execute all tasks in a recipe once. ( Common Entry Point )

      This script dispatches all the tasks in a JSON recipe to handlers in sequence.
      For each task, it calls a subprocess to execute the JSON instructions, waits
      for the process to complete and dispatches the next task, until all tasks are
      complete or a critical failure ( exception ) is raised.

      If an exception is raised in any task, all following tasks are not executed by design.

      Example: python run.py [path to recipe file] --force
      Caution: This script does NOT check if the last job finished, potentially causing overruns.
      Notes:
        - To avoid running the entire script when debugging a single task, the command line
          can easily replace "all" with the name of any "task" in the json.  For example
          python all/run.py project/sample/say_hello.json

        - Can be easily replaced with the following to run only the "hello" task:
          python task/hello/run.py project/sample/say_hello.json

        - Or specified further to run only the second hello task:
          python task/hello/run.py project/sample/say_hello.json -i 2

  """))

  parser.add_argument('json', help='path to tasks json file')

  parser.add_argument(
      '--project',
      '-p',
      help='cloud id of project, defaults to None',
      default=None)
  parser.add_argument(
      '--user',
      '-u',
      help='path to user credentials json file, defaults to GOOGLE_APPLICATION_CREDENTIALS',
      default=None)
  parser.add_argument(
      '--service',
      '-s',
      help='path to service credentials json file, defaults None',
      default=None)
  parser.add_argument(
      '--client',
      '-c',
      help='path to client credentials json file, defaults None',
      default=None)
  parser.add_argument(
      '--key',
      '-k',
      help='Google Cloud API key, also known as developer_key, defaults None',
      default=None)

  parser.add_argument(
      '--verbose',
      '-v',
      help='print all the steps as they happen.',
      action='store_true')
  parser.add_argument(
      '--date',
      '-d',
      help='YYYY-MM-DD format date for which these reports are to be run, default will be today.',
      default='TODAY')
  parser.add_argument(
      '--force',
      '-force',
      help='execute all scripts once then exit.',
      action='store_true')

  parser.add_argument(
      '--trace_print',
      '-tp',
      help='Simplified execution trace of the program written to stdout.',
      action='store_true')
  parser.add_argument(
      '--trace_file',
      '-tf',
      help='Simplified execution trace of the program written to file.',
      action='store_true')

  args = parser.parse_args()

  # load json to get each task
  recipe = get_project(args.json)

  # track per task instance count
  instances = {}

  returncode = EXIT_SUCCESS
  for task in recipe['tasks']:
    script, task = next(iter(task.items()))

    # count instance per task
    instances.setdefault(script, 0)
    instances[script] += 1

    # assemble command ( replace command, use all arguments passed, and add instance )
    command = 'python -W ignore %s/starthinker/task/%s/run.py %s -i %d' % (
      UI_ROOT,
      script,
      ' '.join(sys.argv[1:]),
      instances[script]
    )

    # show command so user can run each task
    print(command)

    # execute command if schedule, return code
    if args.force or is_scheduled(recipe, task):
      if subprocess.Popen(command, shell=True).wait() != EXIT_SUCCESS:
        returncode = EXIT_ERROR

    # skip command if not schedule
    else:
      raise SystemExit(
        'Schedule Skipping: run command manually or add --force to run all'
      )

  # Set lowest return code from all tasks
  exit(returncode)


if __name__ == '__main__':
  main()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

'''
--------------------------------------------------------------

Before running this Airflow module...

  Install StarThinker in cloud composer ( recommended ):

    From Release: pip install starthinker
    From Open Source: pip install git+https://github.com/google/starthinker

  Or push local code to the cloud composer plugins directory ( if pushing local code changes ):

    source install/deploy.sh
    4) Composer Menu
    l) Install All

--------------------------------------------------------------

  If any recipe task has "auth" set to "user" add user credentials:

    1. Ensure an RECIPE['setup']['auth']['user'] = [User Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_user", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/deploy_commandline.md#optional-setup-user-credentials

--------------------------------------------------------------

  If any recipe task has "auth" set to "service" add service credentials:

    1. Ensure an RECIPE['setup']['auth']['service'] = [Service Credentials JSON]

  OR

    1. Visit Airflow UI > Admin > Connections.
    2. Add an Entry called "starthinker_service", fill in the following fields. Last step paste JSON from authentication.
      - Conn Type: Google Cloud Platform
      - Project: Get from https://github.com/google/starthinker/blob/master/tutorials/cloud_project.md
      - Keyfile JSON: Get from: https://github.com/google/starthinker/blob/master/tutorials/cloud_service.md

--------------------------------------------------------------

SDF Download

Download SDF reports into a BigQuery table.

  - Select your filter types and the filter ideas.
  - Enter the <a href='https://developers.google.com/bid-manager/v1.1/sdf/download' target='_blank'>file types</a> using commas.
  - SDF_ will be prefixed to all tables and date appended to daily tables.
  - File types take the following format: FILE_TYPE_CAMPAIGN, FILE_TYPE_AD_GROUP,...

--------------------------------------------------------------

This StarThinker DAG can be extended with any additional tasks from the following sources:
  - https://google.github.io/starthinker/
  - https://github.com/google/starthinker/tree/master/dags

'''

from starthinker.airflow.factory import DAG_Factory

INPUTS = {
  'auth_write': 'service',  # Credentials used for writing data.
  'partner_id': '',  # The sdf file types.
  'file_types': [],  # The sdf file types.
  'filter_type': '',  # The filter type for the filter ids.
  'filter_ids': [],  # Comma separated list of filter ids for the request.
  'dataset': '',  # Dataset to be written to in BigQuery.
  'version': '5',  # The sdf version to be returned.
  'table_suffix': '',  # Optional: Suffix string to put at the end of the table name (Must contain alphanumeric or underscores)
  'time_partitioned_table': False,  # Is the end table a time partitioned
  'create_single_day_table': False,  # Would you like a separate table for each day? This will result in an extra table each day and the end table with the most up to date SDF.
}

RECIPE = {
  'tasks': [
    {
      'dataset': {
        'auth': {
          'field': {
            'name': 'auth_write',
            'kind': 'authentication',
            'order': 1,
            'default': 'service',
            'description': 'Credentials used for writing data.'
          }
        },
        'dataset': {
          'field': {
            'name': 'dataset',
            'kind': 'string',
            'order': 6,
            'default': '',
            'description': 'Dataset to be written to in BigQuery.'
          }
        }
      }
    },
    {
      'sdf': {
        'auth': 'user',
        'version': {
          'field': {
            'name': 'version',
            'kind': 'choice',
            'order': 6,
            'default': '5',
            'description': 'The sdf version to be returned.',
            'choices': [
              'SDF_VERSION_5',
              'SDF_VERSION_5_1'
            ]
          }
        },
        'partner_id': {
          'field': {
            'name': 'partner_id',
            'kind': 'integer',
            'order': 1,
            'description': 'The sdf file types.'
          }
        },
        'file_types': {
          'field': {
            'name': 'file_types',
            'kind': 'string_list',
            'order': 2,
            'default': [
            ],
            'description': 'The sdf file types.'
          }
        },
        'filter_type': {
          'field': {
            'name': 'filter_type',
            'kind': 'choice',
            'order': 3,
            'default': '',
            'description': 'The filter type for the filter ids.',
            'choices': [
              'FILTER_TYPE_ADVERTISER_ID',
              'FILTER_TYPE_CAMPAIGN_ID',
              'FILTER_TYPE_INSERTION_ORDER_ID',
              'FILTER_TYPE_MEDIA_PRODUCT_ID',
              'FILTER_TYPE_LINE_ITEM_ID'
            ]
          }
        },
        'read': {
          'filter_ids': {
            'single_cell': True,
            'values': {
              'field': {
                'name': 'filter_ids',
                'kind': 'integer_list',
                'order': 4,
                'default': [
                ],
                'description': 'Comma separated list of filter ids for the request.'
              }
            }
          }
        },
        'time_partitioned_table': {
          'field': {
            'name': 'time_partitioned_table',
            'kind': 'boolean',
            'order': 7,
            'default': False,
            'description': 'Is the end table a time partitioned'
          }
        },
        'create_single_day_table': {
          'field': {
            'name': 'create_single_day_table',
            'kind': 'boolean',
            'order': 8,
            'default': False,
            'description': 'Would you like a separate table for each day? This will result in an extra table each day and the end table with the most up to date SDF.'
          }
        },
        'dataset': {
          'field': {
            'name': 'dataset',
            'kind': 'string',
            'order': 6,
            'default': '',
            'description': 'Dataset to be written to in BigQuery.'
          }
        },
        'table_suffix': {
          'field': {
            'name': 'table_suffix',
            'kind': 'string',
            'order': 6,
            'default': '',
            'description': 'Optional: Suffix string to put at the end of the table name (Must contain alphanumeric or underscores)'
          }
        }
      }
    }
  ]
}

DAG_FACTORY = DAG_Factory('sdf_to_bigquery', RECIPE, INPUTS)
DAG = DAG_FACTORY.generate()

if __name__ == "__main__":
  DAG_FACTORY.print_commandline()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import json

from starthinker.util.project import project
from starthinker.util.bigquery import rows_to_table, query_to_view
from starthinker.util.dcm import report_build, report_file, report_to_rows, report_clean, report_schema, DCM_CHUNK_SIZE


def write_report(report, dataset, table):
  # turn report file into rows
  rows = report_to_rows(report)
  rows = report_clean(rows)

  if rows:
    if project.verbose:
      print('DYNAMIC COSTS WRITTEN:', table)

    # pull DCM schema automatically
    try:
      schema = report_schema(next(rows))
    except StopIteration:  # report is empty
      raise ValueError('REPORT DID NOT RUN')

    # write report to bigquery
    rows_to_table(project.task['out']['auth'], project.id,
                  project.task['out']['dataset'], table, rows, schema, 0)

  else:
    if project.verbose:
      print('DYNAMIC COSTS REPORT NOT READY:', table)


def report_combos(name, dateRange, schedule, advertiser, campaign,
                  dynamicProfile):
  if project.verbose:
    print('DYNAMIC COSTS COMBOS:', name)

  # basic report schema, with no dynamic elements
  schema = {
      'kind': 'dfareporting#report',
      'type': 'STANDARD',
      'name': 'Dynamic Costs %s - Dynamic Combos ( StarThinker )' % name,
      'schedule': schedule,
      'criteria': {
          'dateRange':
              dateRange,
          'dimensionFilters': [{
              'kind': 'dfareporting#dimensionValue',
              'dimensionName': 'dfa:dynamicProfile',
              'id': dynamicProfile,
              'matchType': 'EXACT'
          }, {
              'kind': 'dfareporting#dimensionValue',
              'dimensionName': 'dfa:advertiser',
              'id': advertiser,
              'matchType': 'EXACT'
          }, {
              'kind': 'dfareporting#dimensionValue',
              'dimensionName': 'dfa:campaign',
              'id': campaign,
              'matchType': 'EXACT'
          }],
          'dimensions': [{
              'kind': 'dfareporting#sortedDimension',
              'name': 'dfa:placement'
          }, {
              'kind': 'dfareporting#sortedDimension',
              'name': 'dfa:placementId'
          }, {
              'kind': 'dfareporting#sortedDimension',
              'name': 'dfa:activity'
          }, {
              'kind': 'dfareporting#sortedDimension',
              'name': 'dfa:activityId'
          }],
          'metricNames': [
              'dfa:impressions', 'dfa:clicks', 'dfa:totalConversions'
          ]
      }
  }

  # add in all reasonable dynamic elements
  for i in range(1, 5 + 1):  # 5 elements/feeds
    for j in range(1, 6 + 1):  # 6 fields per element
      schema['criteria']['dimensions'].append({
          'kind': 'dfareporting#sortedDimension',
          'name': 'dfa:dynamicElement%iField%iValue' % (i, j)
      })

  print(json.dumps(schema, indent=2))

  # create the report if it does not exist
  report = report_build(project.task['auth'], project.task['account'], schema)

  # fetch report file if it exists ( timeout = 0 means grab most reent ready )
  filename, filedata = report_file(project.task['auth'],
                                   project.task['account'], report['id'], None,
                                   60, DCM_CHUNK_SIZE)

  # write report to a table ( avoid collisions as best as possible )
  table_name = 'Dynamic_Costs_%s_Dynamic_Combos' % name
  write_report(filedata, project.task['out']['dataset'], table_name)

  return table_name


def report_main(name, dateRange, schedule, advertiser, campaign, shadow=True):
  if project.verbose:
    print('DYNAMIC COSTS MAIN:', name)

  # base report schema
  schema = {
      'kind': 'dfareporting#report',
      'type': 'STANDARD',
      'name': 'Dynamic Costs %s - Main Advertiser ( StarThinker )' % name,
      'schedule': schedule,
      'criteria': {
          'dateRange': dateRange,
          'dimensionFilters': [{
              'kind': 'dfareporting#dimensionValue',
              'dimensionName': 'dfa:advertiser',
              'id': advertiser,
              'matchType': 'EXACT'
          }, {
              'kind': 'dfareporting#dimensionValue',
              'dimensionName': 'dfa:campaign',
              'id': campaign,
              'matchType': 'EXACT'
          }],
          'dimensions': [{
              'kind': 'dfareporting#sortedDimension',
              'name': 'dfa:placement'
          }, {
              'kind': 'dfareporting#sortedDimension',
              'name': 'dfa:placementId'
          }],
          'metricNames': ['dfa:impressions', 'dfa:clicks']
      }
  }

  # if not using shadow advertiser, pull DBM cost here
  if not shadow:
    schema['criteria']['metricNames'].append('dfa:dbmCost')

  # create the report if it does not exist
  report = report_build(project.task['auth'], project.task['account'], schema)

  # fetch report file if it exists ( timeout = 0 means grab most reent ready )
  filename, filedata = report_file(project.task['auth'],
                                   project.task['account'], report['id'], None,
                                   60, DCM_CHUNK_SIZE)

  # write report to a table ( avoid collisions as best as possible )
  table_name = 'Dynamic_Costs_%s_Main_Advertiser' % name
  write_report(filedata, project.task['out']['dataset'], table_name)

  return table_name


def report_shadow(name, dateRange, schedule, advertiser, campaign):
  if project.verbose:
    print('DYNAMIC COSTS SHADOW:', name)

  # create the report if it does not exist
  report = report_build(
      project.task['auth'], project.task['account'], {
          'kind': 'dfareporting#report',
          'type': 'STANDARD',
          'name': 'Dynamic Costs %s - Shadow Advertiser ( StarThinker )' % name,
          'schedule': schedule,
          'criteria': {
              'dateRange': dateRange,
              'dimensionFilters': [{
                  'dimensionName': 'dfa:advertiser',
                  'id': advertiser,
                  'kind': 'dfareporting#dimensionValue',
                  'matchType': 'EXACT'
              }, {
                  'dimensionName': 'dfa:campaign',
                  'id': campaign,
                  'kind': 'dfareporting#dimensionValue',
                  'matchType': 'EXACT'
              }],
              'dimensions': [{
                  'kind': 'dfareporting#sortedDimension',
                  'name': 'dfa:placement'
              }, {
                  'kind': 'dfareporting#sortedDimension',
                  'name': 'dfa:placementId'
              }],
              'metricNames': ['dfa:dbmCost']
          }
      })

  # fetch report file if it exists ( timeout = 0 means grab most reent ready )
  filename, filedata = report_file(project.task['auth'],
                                   project.task['account'], report['id'])

  # write report to a table ( avoid collisions as best as possible )
  table_name = 'Dynamic_Costs_%s_Shadow_Advertiser' % name
  write_report(filedata, project.task['out']['dataset'], table_name)

  return table_name


def view_combine(name, combos_table, main_table, shadow_table):
  if project.verbose:
    print('DYNAMIC COSTS VIEW:', name)

  if shadow_table:
    query = """
      SELECT
        combos.*,
        (combos.Impressions / main.Impressions) * shadow.Dbm_Cost_Account_Currency AS Scaled_Dbm_Cost_Account_Currency
      FROM `%(project)s.%(dataset)s.%(combos_table)s` combos
      JOIN `%(project)s.%(dataset)s.%(main_table)s` main
      ON combos.Placement_Id = main.Placement_Id
      JOIN `%(project)s.%(dataset)s.%(shadow_table)s` shadow
      ON STARTS_WITH(shadow.Placement, combos.Placement)
      """ % {
          'project': project.id,
          'dataset': project.task['out']['dataset'],
          'combos_table': combos_table,
          'main_table': main_table,
          'shadow_table': shadow_table
      }
  else:
    query = """
      SELECT
        combos.*,
        (combos.Impressions / main.Impressions) * main.Dbm_Cost_Account_Currency AS Scaled_Dbm_Cost_Account_Currency
      FROM `%(project)s.%(dataset)s.%(combos_table)s` combos
      JOIN `%(project)s.%(dataset)s.%(main_table)s` main
      ON combos.Placement_Id = main.Placement_Id
      """ % {
          'project': project.id,
          'dataset': project.task['out']['dataset'],
          'combos_table': combos_table,
          'main_table': main_table
      }

  query_to_view(project.task['out']['auth'], project.id,
                project.task['out']['dataset'],
                'Dynamic_Costs_%s_Analysis' % name, query, False)


@project.from_parameters
def dynamic_costs():

  if project.verbose:
    print('DYNAMIC COSTS PARAMETERS', project.task)

  # allows each advertiser to run multiple reports ( somewhat collision avoidance )
  unique_name = project.task['dynamic_profile_id']

  # check if using wrapped tags
  shadow = project.task['shadow_advertiser_id'] and project.task[
      'shadow_campaign_id']

  # parse date range
  if project.task.get('date_start') and project.task.get('date_end'):
    date_range = {
        'kind': 'dfareporting#dateRange',
        'startDate': project.task['date_start'],
        'endDate': project.task['date_end'],
    }
    schedule = {'active': False, 'repeats': 'DAILY', 'every': 1}
  else:
    date_range = {
        'kind': 'dfareporting#dateRange',
        'relativeDateRange': project.task.get('date_relative', 'YESTERDAY')
    }
    schedule = {'active': True, 'repeats': 'DAILY', 'every': 1}

  combos_table = report_combos(unique_name, date_range, schedule,
                               project.task['main_advertiser_id'],
                               project.task['main_campaign_id'],
                               project.task['dynamic_profile_id'])

  main_table = report_main(unique_name, date_range, schedule,
                           project.task['main_advertiser_id'],
                           project.task['main_campaign_id'], shadow)

  if shadow:
    shadow_table = report_shadow(
        unique_name,
        date_range,
        project.task['shadow_advertiser_id'],
        project.task['shadow_campaign_id'],
    )
  else:
    shadow_table = None

  view_combine(unique_name, combos_table, main_table, shadow_table)


if __name__ == '__main__':
  dynamic_costs()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from starthinker.util.project import project
from starthinker.util.data import put_rows
from starthinker.util.sheets import sheets_read
from starthinker.util.csv import pivot_column_to_row
from starthinker.util.bigquery import table_name_sanitize, query_to_table
from starthinker.util.dcm import report_get, report_delete, report_filter, report_build, report_file, report_to_rows, report_clean, get_account_name, report_schema


def dcm_replicate_accounts():
  if project.verbose:
    print('DCM REPLICATE ACCOUNTS')

  accounts = {}

  # read inputs
  if 'sheet' in project.task['in']:
    rows = sheets_read(
        project.task['auth'],
        project.task['in']['sheet'],
        project.task['in']['tab'],
    )
    # rotate rows so account, advertiser tuple is returned
    return pivot_column_to_row([row[1:] for row in rows])


def dcm_replicate_template():
  if project.verbose:
    print('DCM REPLICATE TEMPLATE')

  report = report_get(
      project.task['auth'],
      project.task['report']['account'],
      project.task['report'].get('id'),
      project.task['report'].get('name'),
  )

  report['criteria']['dimensionFilters'] = []
  del report['id']
  #del report['etag']
  del report['lastModifiedTime']
  del report['ownerProfileId']

  #print(json.dumps(report, indent=2))
  return report


def dcm_replicate_create(account, advertisers, name, template):
  print('DCM REPLICATE CREATE', name)

  # check if report is to be deleted
  if project.task['report'].get('delete', False):
    report_delete(project.task['auth'], account, None, name)

  # add account and advertiser filters ( return new disctionary)
  body = report_filter(project.task['auth'], template, {
      'accountId': {
          'values': account
      },
      'dfa:advertiser': {
          'values': advertisers
      }
  })
  body['name'] = name

  #print('BODY', body)

  # create and run the report if it does not exist
  report = report_build(project.task['auth'], account, body)


def dcm_replicate_download(account, name):

  filename, report = report_file(project.task['auth'], account, None, name)

  if report:
    if project.verbose:
      print('DCM FILE', filename)

    # clean up the report
    rows = report_to_rows(report)
    rows = report_clean(rows)

    # if bigquery, remove header and determine schema
    schema = None
    if 'bigquery' in project.task['out']:
      project.task['out']['bigquery']['table'] = table_name_sanitize(name)
      project.task['out']['bigquery']['schema'] = report_schema(next(rows))
      project.task['out']['bigquery']['skip_rows'] = 0

    # write rows using standard out block in json ( allows customization across all scripts )
    if rows:
      put_rows(project.task['auth'], project.task['out'], rows)


@project.from_parameters
def dcm_replicate():
  if project.verbose:
    print('DCM REPLICATE')

  template = dcm_replicate_template()
  rows = dcm_replicate_accounts()

  # create or update reports
  for row in rows:
    account = row[0]
    advertisers = row[1:]
    name = '%s - Account %s' % (template['name'], account)
    dcm_replicate_create(account, advertisers, name, template)

  # download reports
  for row in rows:
    account = row[0]
    name = '%s - Account %s' % (template['name'], account)
    dcm_replicate_download(account, name)

  # summary table ( combine all data into one table for fast load )
  query_to_table(
      project.task['auth'],
      project.id,
      project.task['out']['bigquery']['dataset'],
      table_name_sanitize('%s - All' % template['name']),
      "SELECT CAST(REPLACE(_TABLE_SUFFIX, '_', '') AS INT64) AS Account_Id, * FROM `%s.%s.%s*`"
      % (project.id, project.task['out']['bigquery']['dataset'],
         table_name_sanitize('%s - Account ' % template['name'])),
      disposition='WRITE_TRUNCATE',
      legacy=False)


if __name__ == '__main__':
  dcm_replicate()



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################
"""Handler that executes { "salesforce":{...}} task in recipe JSON.

This script translates JSON instructions into operations on salesforce
reporting.

"""

from starthinker.util.project import project
from starthinker.util.data import put_rows
from starthinker.util.salesforce import authenticate, query


@project.from_parameters
def salesforce():
  if project.verbose:
    print('Salesforce')

  sf = authenticate(project.task['domain'], project.task['client'],
                    project.task['secret'], project.task['username'],
                    project.task['password'])

  if 'query' in project.task:
    rows = query(sf, project.task['query'])
    if rows:
      put_rows(project.task['auth'], project.task['out'], rows)


if __name__ == '__main__':
  salesforce()



# DEPRECATED: please use starthinker/util/sheets/__init__.py

from starthinker.util.sheets import sheets_clear, sheets_write, sheets_batch_update, sheets_read


def trix_update(auth,
                sheet_id,
                sheet_range,
                data,
                clear=False,
                valueInputOption='RAW',
                retries=10):
  print(
      'WARNING: Deprecated "util.trix.trix_update" function, use "util.sheets.sheets_clear" and "util.sheets.sheets_write" instead.'
  )
  sheet_url = 'https://docs.google.com/spreadsheets/d/%s/edit?usp=sharing' % sheet_id
  sheet_tab, sheet_range = sheet_range.split('!', 1)

  if clear:
    sheets_clear(auth, sheet_tab, sheet_range, sheet_url=sheet_url)
  print('S', sheet_url, sheet_tab, sheet_range, data)
  sheets_write(auth, sheet_tab, sheet_range, data, sheet_url=sheet_url)


def trix_batch_update(auth, sheet_id, data, retries=10):
  print(
      'WARNING: Deprecated "util.trix.trix_batch_update" function, use "util.sheets.sheets_batch_update" instead.'
  )
  sheet_url = 'https://docs.google.com/spreadsheets/d/%s/edit?usp=sharing' % sheet_id
  sheets_batch_update(auth, data, sheet_url=sheet_url)


def trix_read(auth, sheet_id, sheet_range, retries=10):
  print(
      'WARNING: Deprecated "util.trix.trix_read" function, use "util.sheets.sheets_read" instead.'
  )
  sheet_url = 'https://docs.google.com/spreadsheets/d/%s/edit?usp=sharing' % sheet_id
  sheet_tab, sheet_range = sheet_range.split('!', 1)
  sheets_read(auth, sheet_tab, sheet_range, sheet_url=sheet_url)



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################


from starthinker.util.data import get_rows
from starthinker.util.data import put_rows
from starthinker.util.project import project
from starthinker.util.regexp import lookup_id

from starthinker.task.dv_targeter.types import DV360_Targeting
from starthinker.task.dv_targeter.types import TARGETING_TYPES


def advertiser_targeting_load():

  targeting = DV360_Targeting()

  advertisers = [lookup_id(p[0]) for p in get_rows(
    project.task['auth_sheets'], 
    { 'sheets': {
      'sheet': project.task['sheet'],
      'tab': 'Advertisers',
      'range': 'A2:A'
    }}
  )]

  # write to database
  for targeting_type, targeting_meta in TARGETING_TYPES.items():
    if targeting_meta['advertiser']:

      put_rows(
        project.task['auth_bigquery'], 
        { 'bigquery': {
          'dataset': project.task['dataset'],
          'table': 'DV_Targeting_%s' % targeting_meta['name'],
          'schema': targeting.get_schema(targeting_meta['name'], targeting_meta['resource']),
          'disposition':'WRITE_APPEND',
          'format': 'JSON'
        }},
        targeting.get_advertiser(project.task['auth_dv'], advertisers, targeting_type)
      )



from django.conf import settings
from django.core.management.base import BaseCommand, CommandError

from starthinker_worker.log import log_get
from starthinker_ui.recipe.models import Recipe


class Command(BaseCommand):
  help = 'Loads logs into recipes'

  def add_arguments(self, parser):
    parser.add_argument(
        '--recipe',
        action='store',
        dest='recipe',
        default=None,
        help='Run a specific recipe.',
    )

  def handle(self, *args, **kwargs):
    print('Load Recipes')
    recipes = Recipe.objects.filter(
        pk=kwargs['recipe']) if kwargs['recipe'] else Recipe.objects.all()

    print('Load Logs')
    logs = log_get([r.uid() for r in recipes], settings.TIME_ZONE)

    print('Assign Logs')
    for recipe in recipes:
      print('.', end='')
      recipe.set_log(logs.get(recipe.uid(), {}))

    print('\nDone')



###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

import json

from starthinker.util.bigquery import query_to_view
from starthinker.util.bigquery import table_create
from starthinker.util.data import get_rows
from starthinker.util.data import put_rows
from starthinker.util.project import project
from starthinker.util.sheets import sheets_clear

SCHEMA_PREVIEW = [
    {
        'name': 'Operation',
        'type': 'STRING',
        'mode': 'NULLABLE'
    },
    {
        'name': 'Action',
        'type': 'STRING',
        'mode': 'NULLABLE'
    },
    {
        'name': 'Partner',
        'type': 'STRING',
        'mode': 'NULLABLE'
    },
    {
        'name': 'Advertiser',
        'type': 'STRING',
        'mode': 'NULLABLE'
    },
    {
        'name': 'Campaign',
        'type': 'STRING',
        'mode': 'NULLABLE'
    },
    {
        'name': 'Insertion_Order',
        'type': 'STRING',
        'mode': 'NULLABLE'
    },
    {
        'name': 'Line_Item',
        'type': 'STRING',
        'mode': 'NULLABLE'
    },
    {
        'name': 'Parameters',
        'type': 'STRING',
        'mode': 'NULLABLE'
    },
]

SCHEMA_LOG = [
    {
        'name': 'Operation',
        'type': 'STRING',
        'mode': 'NULLABLE'
    },
    {
        'name': 'Action',
        'type': 'STRING',
        'mode': 'NULLABLE'
    },
    {
        'name': 'Partner',
        'type': 'STRING',
        'mode': 'NULLABLE'
    },
    {
        'name': 'Advertiser',
        'type': 'STRING',
        'mode': 'NULLABLE'
    },
    {
        'name': 'Campaign',
        'type': 'STRING',
        'mode': 'NULLABLE'
    },
    {
        'name': 'Insertion_Order',
        'type': 'STRING',
        'mode': 'NULLABLE'
    },
    {
        'name': 'Line_Item',
        'type': 'STRING',
        'mode': 'NULLABLE'
    },
    {
        'name': 'Parameters',
        'type': 'STRING',
        'mode': 'NULLABLE'
    },
    {
        'name': 'Status',
        'type': 'STRING',
        'mode': 'NULLABLE'
    },
    {
        'name': 'Error',
        'type': 'STRING',
        'mode': 'NULLABLE'
    },
]

BUFFER_ERROR = []
BUFFER_SUCCESS = []
BUFFER_LENGTH = 10


def patch_clear():

  table_create(project.task['auth_bigquery'], project.id,
               project.task['dataset'], 'PATCH_Preview', SCHEMA_PREVIEW)

  table_create(project.task['auth_bigquery'], project.id,
               project.task['dataset'], 'PATCH_Log', SCHEMA_LOG)

  sheets_clear(project.task['auth_sheets'], project.task['sheet'], 'Preview', 'A2:Z')

  sheets_clear(project.task['auth_sheets'], project.task['sheet'], 'Error', 'A2:Z')

  sheets_clear(project.task['auth_sheets'], project.task['sheet'], 'Success', 'A2:Z')


def patch_mask(patch):

  def _patch_mask(body):
    mask = set()
    if isinstance(body, dict):
      for parent, value in body.items():
        children = _patch_mask(value)
        if children:
          for child in children:
            mask.add(parent + '.' + child)
        else:
          mask.add(parent)
    elif isinstance(body, (list, tuple)):
      for value in body:
        mask.update(_patch_mask(value))
    return list(mask)

  mask = ','.join(_patch_mask(patch['parameters'].get('body')))
  if mask:
    patch['parameters']['updateMask'] = mask

  return patch


def patch_masks(patches):
  for patch in patches:
    patch_mask(patch)


def patch_log(patch=None):
  global BUFFER_SUCCESS
  global BUFFER_ERROR

  def _patch_write(rows, kind):
    if not rows:
      return

    rows = [(p['operation'], p['action'], p.get('partner'), p.get('advertiser'),
             p.get('campaign'), p.get('insertion_order'), p.get('line_item'),
             json.dumps(p.get('parameters', {}),
                        indent=2), kind, p[kind.lower()]) for p in rows]

    put_rows(
        project.task['auth_bigquery'], {
            'bigquery': {
                'dataset': project.task['dataset'],
                'table': 'PATCH_Log',
                'schema': SCHEMA_LOG,
                'disposition': 'WRITE_APPEND',
                'format': 'CSV'
            }
        }, rows)

    put_rows(
        project.task['auth_sheets'], {
            'sheets': {
                'sheet': project.task['sheet'],
                'tab': kind.title(),
                'range': 'A2',
                'append': True
            }
        }, rows)

  if patch and 'success' in patch:
    BUFFER_SUCCESS.append(patch)
    print('SUCCESS:', patch['success'])
  elif patch and 'error' in patch:
    BUFFER_ERROR.append(patch)
    print('ERROR:', patch['error'])

  if len(BUFFER_SUCCESS) > BUFFER_LENGTH or patch is None:
    _patch_write(BUFFER_SUCCESS, 'SUCCESS')
    BUFFER_SUCCESS = []

  if len(BUFFER_ERROR) > BUFFER_LENGTH or patch is None:
    _patch_write(BUFFER_ERROR, 'ERROR')
    BUFFER_ERROR = []


def patch_preview(patches):
  if patches:
    rows = [(p['operation'], p['action'], p.get('partner'), p.get('advertiser'),
             p.get('campaign'), p.get('insertion_order'), p.get('line_item'),
             json.dumps(p.get('parameters', {}), indent=2)) for p in patches]

    put_rows(
        project.task['auth_bigquery'], {
            'bigquery': {
                'dataset': project.task['dataset'],
                'table': 'PATCH_Preview',
                'schema': SCHEMA_PREVIEW,
                'disposition': 'WRITE_APPEND',
                'format': 'CSV'
            }
        }, rows)

    put_rows(
        project.task['auth_sheets'], {
            'sheets': {
                'sheet': project.task['sheet'],
                'tab': 'Preview',
                'range': 'A2',
            }
        }, rows)



# -*- coding: utf-8 -*-

###########################################################################
#
#  Copyright 2020 Google LLC
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
###########################################################################

from django.shortcuts import render
from django.http import HttpResponse, HttpResponseRedirect
from django.contrib import messages

from starthinker_ui.account.decorators import permission_admin
from starthinker_ui.project.models import Project
from starthinker_ui.project.forms import ProjectForm


def project_list(request):
  projects = request.user.project_set.all(
  ) if request.user.is_authenticated else None
  return render(request, 'project/project_list.html', {'projects': projects})


@permission_admin()
def project_edit(request, pk=None):
  project = request.user.project_set.get(pk=pk) if pk else None

  if request.method == 'POST':
    form_project = ProjectForm(request.user, request.POST, instance=project)
    if form_project.is_valid():
      form_project.save()
      messages.success(request, 'Project updated.')
      return HttpResponseRedirect(form_project.instance.link_edit())
    else:
      print('ERRORS', form_project.get_errors())
      messages.error(request, 'Project Errors: %s' % form_project.get_errors())
  else:
    form_project = ProjectForm(request.user, instance=project)

  return render(request, 'project/project_edit.html',
                {'form_project': form_project})


@permission_admin()
def project_delete(request, pk=None):
  request.user.project_set.filter(pk=pk).delete()
  messages.success(request, 'Project deleted.')
  return HttpResponseRedirect('/project/')
